import os
import re
from typing import List

import boto3
import pandas as pd
import pytest
from moto import mock_s3

from great_expectations.core import IDDict
from great_expectations.core.batch import BatchDefinition
from great_expectations.datasource.data_connector.util import (
    sanitize_prefix,
    sanitize_prefix_for_s3,
)
from great_expectations.experimental.datasources.data_asset.data_connector import (
    DataConnector,
    S3DataConnector,
)
from great_expectations.experimental.datasources.interfaces import BatchRequest


@pytest.mark.integration
@mock_s3
def test_basic_instantiation():
    region_name: str = "us-east-1"
    bucket: str = "test_bucket"
    conn = boto3.resource("s3", region_name=region_name)
    conn.create_bucket(Bucket=bucket)
    client = boto3.client("s3", region_name=region_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    for key in keys:
        client.put_object(
            Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
        )

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"alpha-(.*)\.csv"),
        prefix="",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    # Missing "data_asset_name" argument.
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="something",
                options={},
            )
        )


@pytest.mark.integration
@mock_s3
def test_instantiation_batching_regex_does_not_match_paths():
    region_name: str = "us-east-1"
    bucket: str = "test_bucket"
    conn = boto3.resource("s3", region_name=region_name)
    conn.create_bucket(Bucket=bucket)
    client = boto3.client("s3", region_name=region_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    for key in keys:
        client.put_object(
            Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
        )

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"beta-(.*)\.csv"),
        prefix="",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert len(my_data_connector.get_unmatched_data_references()) == 3


@pytest.mark.integration
@mock_s3
def test_return_all_batch_definitions_unsorted():
    region_name: str = "us-east-1"
    bucket: str = "test_bucket"
    conn = boto3.resource("s3", region_name=region_name)
    conn.create_bucket(Bucket=bucket)
    client = boto3.client("s3", region_name=region_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "alex_20200809_1000.csv",
        "eugene_20200809_1500.csv",
        "james_20200811_1009.csv",
        "abe_20200809_1040.csv",
        "will_20200809_1002.csv",
        "james_20200713_1567.csv",
        "eugene_20201129_1900.csv",
        "will_20200810_1001.csv",
        "james_20200810_1003.csv",
        "alex_20200819_1300.csv",
    ]
    for key in keys:
        client.put_object(
            Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
        )

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.+)\.csv"),
        prefix="",
    )
    # with missing BatchRequest arguments
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list()

    # with empty options
    unsorted_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_dataframe_datasource",
            data_asset_name="my_s3_data_asset",
            options={},
        )
    )
    expected: List[BatchDefinition] = [
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "abe", "timestamp": "20200809", "price": "1040"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "alex", "timestamp": "20200809", "price": "1000"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "alex", "timestamp": "20200819", "price": "1300"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "eugene", "timestamp": "20200809", "price": "1500"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "eugene", "timestamp": "20201129", "price": "1900"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "james", "timestamp": "20200713", "price": "1567"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "james", "timestamp": "20200810", "price": "1003"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "james", "timestamp": "20200811", "price": "1009"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "will", "timestamp": "20200809", "price": "1002"}
            ),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict(
                {"name": "will", "timestamp": "20200810", "price": "1001"}
            ),
        ),
    ]
    assert expected == unsorted_batch_definition_list

    # with specified Batch query options
    unsorted_batch_definition_list = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_dataframe_datasource",
            data_asset_name="my_s3_data_asset",
            options={"name": "alex", "timestamp": "20200819", "price": "1300"},
        )
    )
    assert expected[2:3] == unsorted_batch_definition_list


# TODO: <Alex>ALEX-UNCOMMENT_WHEN_SORTERS_ARE_INCLUDED_AND_TEST_SORTED_BATCH_DEFINITION_LIST</Alex>
# TODO: <Alex>ALEX</Alex>
# @pytest.mark.integration
# @mock_s3
# def test_return_all_batch_definitions_sorted():
#     region_name: str = "us-east-1"
#     bucket: str = "test_bucket"
#     conn = boto3.resource("s3", region_name=region_name)
#     conn.create_bucket(Bucket=bucket)
#     client = boto3.client("s3", region_name=region_name)
#
#     test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})
#
#     keys: List[str] = [
#         "alex_20200809_1000.csv",
#         "eugene_20200809_1500.csv",
#         "james_20200811_1009.csv",
#         "abe_20200809_1040.csv",
#         "will_20200809_1002.csv",
#         "james_20200713_1567.csv",
#         "eugene_20201129_1900.csv",
#         "will_20200810_1001.csv",
#         "james_20200810_1003.csv",
#         "alex_20200819_1300.csv",
#     ]
#     for key in keys:
#         client.put_object(
#             Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
#         )
#
#     my_data_connector: DataConnector = S3DataConnector(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_s3_data_asset",
#         bucket=bucket,
#         batching_regex=re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.+)\.csv"),
#         prefix="",
#     )
#     # noinspection PyProtectedMember
#     my_data_connector._get_data_references_cache()
#
#     assert my_data_connector.get_data_reference_count() == 3
#     assert my_data_connector._get_data_reference_list()[:3] == [
#         "alpha-1.csv",
#         "alpha-2.csv",
#         "alpha-3.csv",
#     ]
#     assert my_data_connector.get_unmatched_data_references()[:3] == [
#         "alpha-1.csv",
#         "alpha-2.csv",
#         "alpha-3.csv",
#     ]
#     assert len(my_data_connector.get_unmatched_data_references()) == 3
#
#     sorted_batch_definition_list: List[BatchDefinition] = (
#         my_data_connector.get_batch_definition_list(
#             BatchRequest(
#                 datasource_name="my_dataframe_datasource",
#                 data_asset_name="my_s3_data_asset",
#                 options={},
#             )
#         )
#     )
#
#     expected: List[BatchDefinition] = [
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "abe", "timestamp": "20200809", "price": "1040"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "alex", "timestamp": "20200819", "price": "1300"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "alex", "timestamp": "20200809", "price": "1000"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "eugene", "timestamp": "20201129", "price": "1900"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "eugene", "timestamp": "20200809", "price": "1500"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "james", "timestamp": "20200811", "price": "1009"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "james", "timestamp": "20200810", "price": "1003"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "james", "timestamp": "20200713", "price": "1567"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "will", "timestamp": "20200810", "price": "1001"}
#             ),
#         ),
#         BatchDefinition(
#             datasource_name="my_dataframe_datasource",
#             data_connector_name="experimental",
#             data_asset_name="my_s3_data_asset",
#             batch_identifiers=IDDict(
#                 {"name": "will", "timestamp": "20200809", "price": "1002"}
#             ),
#         ),
#     ]
#
#     # TEST 1: Sorting works
#     assert expected == sorted_batch_definition_list
#
#     my_batch_request: BatchRequest = BatchRequest(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_s3_data_asset",
#         options={
#             "name": "james",
#             "timestamp": "20200713",
#             "price": "1567",
#         },
#     )
#
#     my_batch_definition_list: List[BatchDefinition]
#     my_batch_definition: BatchDefinition
#
#     # TEST 2: Should only return the specified partition
#     my_batch_definition_list = (
#         my_data_connector.get_batch_definition_list(
#             batch_request=my_batch_request
#         )
#     )
#     assert len(my_batch_definition_list) == 1
#     my_batch_definition = my_batch_definition_list[0]
#
#     expected_batch_definition = BatchDefinition(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_s3_data_asset",
#         batch_identifiers={
#             "name": "james",
#             "timestamp": "20200713",
#             "price": "1567",
#         },
#     )
#     assert my_batch_definition == expected_batch_definition
#
#     # TEST 3: Without BatchRequest (query) options, should return all 10
#     my_batch_request: BatchRequest = BatchRequest(
#         datasource_name="my_dataframe_datasource",
#         data_asset_name="my_s3_data_asset",
#         options={},
#     )
#     # should return 10
#     my_batch_definition_list = (
#         my_data_connector.get_batch_definition_list(
#             batch_request=my_batch_request
#         )
#     )
#     assert len(my_batch_definition_list) == 10
# TODO: <Alex>ALEX</Alex>


@pytest.mark.integration
@mock_s3
def test_return_only_unique_batch_definitions():
    region_name: str = "us-east-1"
    bucket: str = "test_bucket"
    conn = boto3.resource("s3", region_name=region_name)
    conn.create_bucket(Bucket=bucket)
    client = boto3.client("s3", region_name=region_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
        "B/file_1.csv",
        "B/file_2.csv",
    ]
    for key in keys:
        client.put_object(
            Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
        )

    my_data_connector: DataConnector

    my_data_connector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<name>.+)/.+\.csv"),
        prefix="A",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    expected: List[BatchDefinition] = [
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict({"directory": "B", "filename": "file_1.csv"}),
        ),
        BatchDefinition(
            datasource_name="my_dataframe_datasource",
            data_connector_name="experimental",
            data_asset_name="my_s3_data_asset",
            batch_identifiers=IDDict({"directory": "B", "filename": "file_2.csv"}),
        ),
    ]

    my_data_connector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<directory>.+)/(?P<filename>.+\.csv)"),
        prefix="B",
    )

    unsorted_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_dataframe_datasource",
            data_asset_name="my_s3_data_asset",
            options={},
        )
    )
    assert expected == unsorted_batch_definition_list


@pytest.mark.integration
@mock_s3
def test_alpha():
    region_name: str = "us-east-1"
    bucket: str = "test_bucket"
    conn = boto3.resource("s3", region_name=region_name)
    conn.create_bucket(Bucket=bucket)
    client = boto3.client("s3", region_name=region_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
        "test_dir_alpha/D.csv",
    ]
    for key in keys:
        client.put_object(
            Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
        )

    my_data_connector: DataConnector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<part_1>.+)\.csv"),
        prefix="test_dir_alpha",
    )
    assert my_data_connector.get_data_reference_count() == 4
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_batch_definition_list: List[BatchDefinition]
    my_batch_definition: BatchDefinition

    my_batch_request: BatchRequest

    # Try to fetch a batch from a nonexistent asset
    my_batch_request = BatchRequest(
        datasource_name="BASE", data_asset_name="A", options={}
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        options={"part_1": "test_dir_alpha/B"},
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 1


@pytest.mark.integration
@mock_s3
def test_foxtrot():
    region_name: str = "us-east-1"
    bucket: str = "test_bucket"
    conn = boto3.resource("s3", region_name=region_name)
    conn.create_bucket(Bucket=bucket)
    client = boto3.client("s3", region_name=region_name)

    test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})

    keys: List[str] = [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
        "test_dir_foxtrot/D/D-aaa.csv",
        "test_dir_foxtrot/D/D-bbb.csv",
        "test_dir_foxtrot/D/D-ccc.csv",
        "test_dir_foxtrot/D/D-ddd.csv",
        "test_dir_foxtrot/D/D-eee.csv",
    ]
    for key in keys:
        client.put_object(
            Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
        )

    my_data_connector: DataConnector

    my_data_connector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv"),
        prefix="",
    )
    assert my_data_connector.get_data_reference_count() == 0
    assert my_data_connector.get_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_data_connector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv"),
        prefix="test_dir_foxtrot/A",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_data_connector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.txt"),
        prefix="test_dir_foxtrot/B",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_data_connector = S3DataConnector(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        bucket=bucket,
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.+)\.csv"),
        prefix="test_dir_foxtrot/C",
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert len(my_data_connector.get_unmatched_data_references()) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_dataframe_datasource",
        data_asset_name="my_s3_data_asset",
        options={},
    )
    my_batch_definition_list: List[
        BatchDefinition
    ] = my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    assert len(my_batch_definition_list) == 3


# TODO: <Alex>ALEX-UNCOMMENT_WHEN_SORTERS_ARE_INCLUDED_AND_TEST_SORTED_BATCH_DEFINITION_LIST</Alex>
# TODO: <Alex>ALEX</Alex>
# @pytest.mark.integration
# @mock_s3
# def test_return_all_batch_definitions_sorted_sorter_named_that_does_not_match_group(
#     tmp_path_factory,
# ):
#     region_name: str = "us-east-1"
#     bucket: str = "test_bucket"
#     conn = boto3.resource("s3", region_name=region_name)
#     conn.create_bucket(Bucket=bucket)
#     client = boto3.client("s3", region_name=region_name)
#
#     test_df: pd.DataFrame = pd.DataFrame(data={"col1": [1, 2], "col2": [3, 4]})
#
#     keys: List[str] = [
#         "alex_20200809_1000.csv",
#         "eugene_20200809_1500.csv",
#         "james_20200811_1009.csv",
#         "abe_20200809_1040.csv",
#         "will_20200809_1002.csv",
#         "james_20200713_1567.csv",
#         "eugene_20201129_1900.csv",
#         "will_20200810_1001.csv",
#         "james_20200810_1003.csv",
#         "alex_20200819_1300.csv",
#     ]
#     for key in keys:
#         client.put_object(
#             Bucket=bucket, Body=test_df.to_csv(index=False).encode("utf-8"), Key=key
#         )
#
#     my_data_connector_yaml = yaml.load(
#         f"""
#         class_name: S3DataConnector
#         datasource_name: test_environment
#         base_directory: {base_directory}
#         glob_directive: "*.csv"
#         assets:
#             my_s3_data_asset:
#                 pattern: (.+)_(.+)_(.+)\\.csv
#                 group_names:
#                     - name
#                     - timestamp
#                     - price
#         default_regex:
#             pattern: (.+)_.+_.+\\.csv
#             group_names:
#                 - name
#         sorters:
#             - orderby: asc
#               class_name: LexicographicSorter
#               name: name
#             - datetime_format: "%Y%m%d"
#               orderby: desc
#               class_name: DateTimeSorter
#               name: timestamp
#             - orderby: desc
#               class_name: NumericSorter
#               name: for_me_Me_Me
#     """,
#     )
#     with pytest.raises(gx_exceptions.DataConnectorError):
#         # noinspection PyUnusedLocal
#         my_data_connector: S3DataConnector = (
#             instantiate_class_from_config(
#                 config=my_data_connector_yaml,
#                 runtime_environment={
#                     "name": "experimental",
#                     "execution_engine": PandasExecutionEngine(),
#                 },
#                 config_defaults={
#                     "module_name": "great_expectations.datasource.data_connector"
#                 },
#             )
#         )
# TODO: <Alex>ALEX</Alex>


def test_sanitize_prefix_behaves_the_same_as_local_files():
    def check_sameness(prefix, expected_output):
        s3_sanitized = sanitize_prefix_for_s3(prefix)
        file_system_sanitized = sanitize_prefix(prefix)
        if os.sep == "\\":  # Fix to ensure tests work on Windows
            file_system_sanitized = file_system_sanitized.replace("\\", "/")

        assert file_system_sanitized == expected_output, (
            f"Expected output does not match original sanitization behavior, got "
            f"{file_system_sanitized} instead of {expected_output}"
        )
        assert (
            s3_sanitized == expected_output == file_system_sanitized
        ), f'S3 sanitized result is incorrect, "{s3_sanitized} instead of {expected_output}'

    # Copy of all samples from tests/datasource/data_connector/test_file_path_data_connector.py
    check_sameness("foo/", "foo/")
    check_sameness("bar", "bar/")
    check_sameness("baz.txt", "baz.txt")
    check_sameness("a/b/c/baz.txt", "a/b/c/baz.txt")

    # A couple additional checks
    check_sameness("a/b/c", "a/b/c/")
    check_sameness("a.x/b/c", "a.x/b/c/")
    check_sameness("path/to/folder.something/", "path/to/folder.something/")
    check_sameness("path/to/folder.something", "path/to/folder.something")



from __future__ import annotations

import copy
import json
import logging
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)

import jsonpatch
from marshmallow import Schema, ValidationError, fields, post_dump, post_load
from pyparsing import ParseResults
from typing_extensions import TypedDict

from great_expectations.alias_types import JSONValues  # noqa: TCH001
from great_expectations.core._docs_decorators import new_argument, public_api
from great_expectations.core.evaluation_parameters import (
    _deduplicate_evaluation_parameter_dependencies,
    build_evaluation_parameters,
    find_evaluation_parameter_dependencies,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.urn import ge_urn
from great_expectations.core.util import (
    convert_to_json_serializable,
    ensure_json_serializable,
    nested_update,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.exceptions import (
    ClassInstantiationError,
    ExpectationNotFoundError,
    InvalidExpectationConfigurationError,
    InvalidExpectationKwargsError,
    ParserError,
)
from great_expectations.expectations.registry import get_expectation_impl
from great_expectations.render import RenderedAtomicContent, RenderedAtomicContentSchema
from great_expectations.types import SerializableDictDot

if TYPE_CHECKING:
    from great_expectations.core import ExpectationValidationResult
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation import Expectation
    from great_expectations.render.renderer.inline_renderer import InlineRendererConfig
    from great_expectations.rule_based_profiler.config import RuleBasedProfilerConfig
    from great_expectations.validator.validator import Validator
logger = logging.getLogger(__name__)


def parse_result_format(result_format: Union[str, dict]) -> dict:
    """This is a simple helper utility that can be used to parse a string result_format into the dict format used
    internally by great_expectations. It is not necessary but allows shorthand for result_format in cases where
    there is no need to specify a custom partial_unexpected_count."""
    if isinstance(result_format, str):
        result_format = {
            "result_format": result_format,
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        }
    else:
        if (
            "include_unexpected_rows" in result_format
            and "result_format" not in result_format
        ):
            raise ValueError(
                "When using `include_unexpected_rows`, `result_format` must be explicitly specified"
            )

        if "partial_unexpected_count" not in result_format:
            result_format["partial_unexpected_count"] = 20

        if "include_unexpected_rows" not in result_format:
            result_format["include_unexpected_rows"] = False

    return result_format


class ExpectationContext(SerializableDictDot):
    def __init__(self, description: Optional[str] = None) -> None:
        self._description = description

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value) -> None:
        self._description = value


class ExpectationContextSchema(Schema):
    description = fields.String(required=False, allow_none=True)

    @post_load
    def make_expectation_context(self, data, **kwargs):
        return ExpectationContext(**data)


class KWargDetailsDict(TypedDict):
    domain_kwargs: tuple[str, ...]
    success_kwargs: tuple[str, ...]
    default_kwarg_values: dict[str, str | bool | float | RuleBasedProfilerConfig | None]


@public_api
@new_argument(
    argument_name="rendered_content",
    version="0.15.14",
    message="Used to include rendered content dictionary in expectation configuration.",
)
@new_argument(
    argument_name="ge_cloud_id",
    version="0.13.36",
    message="Used in GX Cloud deployments.",
)
@new_argument(
    argument_name="expectation_context",
    version="0.13.44",
    message="Used to support column descriptions in GX Cloud.",
)
class ExpectationConfiguration(SerializableDictDot):
    """Denies the parameters and name of a specific expectation.

    Args:
        expectation_type: The name of the expectation class to use in snake case, e.g. `expect_column_values_to_not_be_null`.
        kwargs: The keyword arguments to pass to the expectation class.
        meta: A dictionary of metadata to attach to the expectation.
        success_on_last_run: Whether the expectation succeeded on the last run.
        ge_cloud_id: The corresponding GX Cloud ID for the expectation.
        expectation_context: The context for the expectation.
        rendered_content: Rendered content for the expectation.
    Raises:
        InvalidExpectationConfigurationError: If `expectation_type` arg is not a str.
        InvalidExpectationConfigurationError: If `kwargs` arg is not a dict.
        InvalidExpectationKwargsError: If domain kwargs are missing.
        ValueError: If a `domain_type` cannot be determined.
    """

    kwarg_lookup_dict: ClassVar[Mapping[str, KWargDetailsDict]] = {
        "expect_column_to_exist": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("column_index",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "column_index": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_table_columns_to_match_ordered_list": {
            "domain_kwargs": (),
            "success_kwargs": ("column_list",),
            "default_kwarg_values": {
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_table_columns_to_match_set": {
            "domain_kwargs": (),
            "success_kwargs": ("column_set", "exact_match"),
            "default_kwarg_values": {
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "exact_match": True,
            },
        },
        "expect_table_column_count_to_be_between": {
            "domain_kwargs": (),
            "success_kwargs": ("min_value", "max_value"),
            "default_kwarg_values": {
                "min_value": None,
                "max_value": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_table_column_count_to_equal": {
            "domain_kwargs": (),
            "success_kwargs": ("value",),
            "default_kwarg_values": {
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_table_row_count_to_be_between": {
            "domain_kwargs": (),
            "success_kwargs": ("min_value", "max_value"),
            "default_kwarg_values": {
                "min_value": None,
                "max_value": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_table_row_count_to_equal": {
            "domain_kwargs": (),
            "success_kwargs": ("value",),
            "default_kwarg_values": {
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_unique": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("mostly",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_not_be_null": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("mostly",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_compound_columns_to_be_unique": {
            "domain_kwargs": ("column_list", "row_condition", "condition_parser"),
            "success_kwargs": ("ignore_row_if",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "ignore_row_if": "all_values_are_missing",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_null": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("mostly",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_of_type": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("type_", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_in_type_list": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("type_list", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_in_set": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value_set", "mostly", "parse_strings_as_datetimes"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "parse_strings_as_datetimes": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_not_be_in_set": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value_set", "mostly", "parse_strings_as_datetimes"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "parse_strings_as_datetimes": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": (
                "min_value",
                "max_value",
                "strict_min",
                "strict_max",
                "allow_cross_type_comparisons",
                "parse_strings_as_datetimes",
                "output_strftime_format",
                "mostly",
            ),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "allow_cross_type_comparisons": None,
                "parse_strings_as_datetimes": None,
                "output_strftime_format": None,
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_increasing": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("strictly", "parse_strings_as_datetimes", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "strictly": None,
                "parse_strings_as_datetimes": None,
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_decreasing": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("strictly", "parse_strings_as_datetimes", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "strictly": None,
                "parse_strings_as_datetimes": None,
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_value_lengths_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_value_lengths_to_equal": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_match_regex": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("regex", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_not_match_regex": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("regex", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_match_regex_list": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("regex_list", "match_on", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "match_on": "any",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_not_match_regex_list": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("regex_list", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_match_strftime_format": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("strftime_format", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_dateutil_parseable": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("mostly",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_be_json_parseable": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("mostly",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_values_to_match_json_schema": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("json_schema", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("distribution", "p_value", "params"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "p_value": 0.05,
                "params": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_distinct_values_to_be_in_set": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value_set", "parse_strings_as_datetimes"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "parse_strings_as_datetimes": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_distinct_values_to_equal_set": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value_set", "parse_strings_as_datetimes"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "parse_strings_as_datetimes": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_distinct_values_to_contain_set": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value_set", "parse_strings_as_datetimes"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "parse_strings_as_datetimes": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_mean_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value", "strict_min", "strict_max"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_median_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value", "strict_min", "strict_max"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_quantile_values_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("quantile_ranges", "allow_relative_error"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "allow_relative_error": False,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_stdev_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value", "strict_min", "strict_max"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_unique_value_count_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_proportion_of_unique_values_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value", "strict_min", "strict_max"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_most_common_value_to_be_in_set": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("value_set", "ties_okay"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "ties_okay": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_sum_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("min_value", "max_value", "strict_min", "strict_max"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_min_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": (
                "min_value",
                "max_value",
                "strict_min",
                "strict_max",
                "parse_strings_as_datetimes",
                "output_strftime_format",
            ),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "parse_strings_as_datetimes": None,
                "output_strftime_format": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_max_to_be_between": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": (
                "min_value",
                "max_value",
                "strict_min",
                "strict_max",
                "parse_strings_as_datetimes",
                "output_strftime_format",
            ),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "min_value": None,
                "max_value": None,
                "strict_min": False,
                "strict_max": False,
                "parse_strings_as_datetimes": None,
                "output_strftime_format": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_chisquare_test_p_value_to_be_greater_than": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("partition_object", "p", "tail_weight_holdout"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "partition_object": None,
                "p": 0.05,
                "tail_weight_holdout": 0,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_bootstrapped_ks_test_p_value_to_be_greater_than": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": (
                "partition_object",
                "p",
                "bootstrap_samples",
                "bootstrap_sample_size",
            ),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "partition_object": None,
                "p": 0.05,
                "bootstrap_samples": None,
                "bootstrap_sample_size": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_kl_divergence_to_be_less_than": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": (
                "partition_object",
                "threshold",
                "tail_weight_holdout",
                "internal_weight_holdout",
                "bucketize_data",
            ),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "partition_object": None,
                "threshold": None,
                "tail_weight_holdout": 0,
                "internal_weight_holdout": 0,
                "bucketize_data": True,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_pair_values_to_be_equal": {
            "domain_kwargs": (
                "column_A",
                "column_B",
                "row_condition",
                "condition_parser",
            ),
            "success_kwargs": ("ignore_row_if",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "ignore_row_if": "both_values_are_missing",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_pair_values_A_to_be_greater_than_B": {
            "domain_kwargs": (
                "column_A",
                "column_B",
                "row_condition",
                "condition_parser",
            ),
            "success_kwargs": (
                "or_equal",
                "parse_strings_as_datetimes",
                "allow_cross_type_comparisons",
                "ignore_row_if",
            ),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "or_equal": None,
                "parse_strings_as_datetimes": None,
                "allow_cross_type_comparisons": None,
                "ignore_row_if": "both_values_are_missing",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_pair_values_to_be_in_set": {
            "domain_kwargs": (
                "column_A",
                "column_B",
                "row_condition",
                "condition_parser",
            ),
            "success_kwargs": ("value_pairs_set", "ignore_row_if", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "ignore_row_if": "both_values_are_missing",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_multicolumn_values_to_be_unique": {
            "domain_kwargs": ("column_list", "row_condition", "condition_parser"),
            "success_kwargs": ("ignore_row_if",),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "ignore_row_if": "all_values_are_missing",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_multicolumn_sum_to_equal": {
            "domain_kwargs": ("column_list", "row_condition", "condition_parser"),
            "success_kwargs": ("sum_total", "ignore_row_if"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "ignore_row_if": "all_values_are_missing",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "_expect_column_values_to_be_of_type__aggregate": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("type_", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "_expect_column_values_to_be_of_type__map": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("type_", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "_expect_column_values_to_be_in_type_list__aggregate": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("type_list", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "_expect_column_values_to_be_in_type_list__map": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("type_list", "mostly"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_column_value_z_scores_to_be_less_than": {
            "domain_kwargs": ("column", "row_condition", "condition_parser"),
            "success_kwargs": ("threshold", "mostly", "double_sided"),
            "default_kwarg_values": {
                "row_condition": None,
                "condition_parser": "pandas",
                "mostly": 1,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
            },
        },
        "expect_file_line_regex_match_count_to_be_between": {
            "domain_kwargs": (),
            "success_kwargs": (
                "regex",
                "expected_min_count",
                "expected_max_count",
                "skip",
            ),
            "default_kwarg_values": {
                "expected_min_count": 0,
                "expected_max_count": None,
                "skip": None,
                "mostly": 1,
                "nonnull_lines_regex": r"^\s*$",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
                "_lines": None,
            },
        },
        "expect_file_line_regex_match_count_to_equal": {
            "domain_kwargs": (),
            "success_kwargs": ("regex", "expected_count", "skip"),
            "default_kwarg_values": {
                "expected_count": 0,
                "skip": None,
                "mostly": 1,
                "nonnull_lines_regex": r"^\s*$",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
                "_lines": None,
            },
        },
        "expect_file_hash_to_equal": {
            "domain_kwargs": (),
            "success_kwargs": ("value", "hash_alg"),
            "default_kwarg_values": {
                "hash_alg": "md5",
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
            },
        },
        "expect_file_size_to_be_between": {
            "domain_kwargs": (),
            "success_kwargs": ("minsize", "maxsize"),
            "default_kwarg_values": {
                "minsize": 0,
                "maxsize": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
            },
        },
        "expect_file_to_exist": {
            "domain_kwargs": (),
            "success_kwargs": ("filepath",),
            "default_kwarg_values": {
                "filepath": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
            },
        },
        "expect_file_to_have_valid_table_header": {
            "domain_kwargs": (),
            "success_kwargs": ("regex", "skip"),
            "default_kwarg_values": {
                "skip": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
            },
        },
        "expect_file_to_be_valid_json": {
            "domain_kwargs": (),
            "success_kwargs": ("schema",),
            "default_kwarg_values": {
                "schema": None,
                "result_format": "BASIC",
                "include_config": True,
                "catch_exceptions": False,
                "meta": None,
            },
        },
    }

    runtime_kwargs: ClassVar[tuple[str, ...]] = (
        "result_format",
        "include_config",
        "catch_exceptions",
    )

    def __init__(
        self,
        expectation_type: str,
        kwargs: dict,
        meta: Optional[dict] = None,
        success_on_last_run: Optional[bool] = None,
        ge_cloud_id: Optional[str] = None,
        expectation_context: Optional[ExpectationContext] = None,
        rendered_content: Optional[List[RenderedAtomicContent]] = None,
    ) -> None:
        if not isinstance(expectation_type, str):
            raise InvalidExpectationConfigurationError(
                "expectation_type must be a string"
            )
        self._expectation_type = expectation_type
        if not isinstance(kwargs, dict):
            raise InvalidExpectationConfigurationError(
                "expectation configuration kwargs must be a dict."
            )
        self._kwargs = kwargs
        # the kwargs before evaluation parameters are evaluated
        self._raw_kwargs: dict[str, Any] | None = None
        if meta is None:
            meta = {}
        # We require meta information to be serializable, but do not convert until necessary
        ensure_json_serializable(meta)
        self.meta = meta
        self.success_on_last_run = success_on_last_run
        self._ge_cloud_id = ge_cloud_id
        self._expectation_context = expectation_context
        self._rendered_content = rendered_content

    def process_evaluation_parameters(
        self,
        evaluation_parameters,
        interactive_evaluation: bool = True,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        if not self._raw_kwargs:
            evaluation_args, _ = build_evaluation_parameters(
                expectation_args=self._kwargs,
                evaluation_parameters=evaluation_parameters,
                interactive_evaluation=interactive_evaluation,
                data_context=data_context,
            )

            self._raw_kwargs = self._kwargs
            self._kwargs = evaluation_args
        else:
            logger.debug(
                "evaluation_parameters have already been built on this expectation"
            )

    def get_raw_configuration(self) -> ExpectationConfiguration:
        # return configuration without substituted evaluation parameters
        raw_config = deepcopy(self)
        if raw_config._raw_kwargs is not None:
            raw_config._kwargs = raw_config._raw_kwargs
            raw_config._raw_kwargs = None

        return raw_config

    def patch(self, op: str, path: str, value: Any) -> ExpectationConfiguration:
        """

        Args:
            op: A jsonpatch operation. One of 'add', 'replace', or 'remove'
            path: A jsonpatch path for the patch operation
            value: The value to patch

        Returns:
            The patched ExpectationConfiguration object
        """
        if op not in ["add", "replace", "remove"]:
            raise ValueError("Op must be either 'add', 'replace', or 'remove'")

        try:
            valid_path = path.split("/")[1]
        except IndexError:
            raise IndexError(
                "Ensure you have a valid jsonpatch path of the form '/path/foo' "
                "(see http://jsonpatch.com/)"
            )

        if valid_path not in self.get_runtime_kwargs().keys():
            raise ValueError("Path not available in kwargs (see http://jsonpatch.com/)")

        # TODO: Call validate_kwargs when implemented
        patch = jsonpatch.JsonPatch([{"op": op, "path": path, "value": value}])

        patch.apply(self.kwargs, in_place=True)
        return self

    @property
    def ge_cloud_id(self) -> Optional[str]:
        return self._ge_cloud_id

    @ge_cloud_id.setter
    def ge_cloud_id(self, value: str) -> None:
        self._ge_cloud_id = value

    @property
    def expectation_context(self) -> Optional[ExpectationContext]:
        return self._expectation_context

    @expectation_context.setter
    def expectation_context(self, value: ExpectationContext) -> None:
        self._expectation_context = value

    @property
    def expectation_type(self) -> str:
        return self._expectation_type

    @property
    def kwargs(self) -> dict:
        return self._kwargs

    @kwargs.setter
    def kwargs(self, value: dict) -> None:
        self._kwargs = value

    @property
    def rendered_content(self) -> Optional[List[RenderedAtomicContent]]:
        return self._rendered_content

    @rendered_content.setter
    def rendered_content(self, value: Optional[List[RenderedAtomicContent]]) -> None:
        self._rendered_content = value

    def _get_default_custom_kwargs(self) -> KWargDetailsDict:
        # NOTE: this is a holdover until class-first expectations control their
        # defaults, and so defaults are inherited.
        if self.expectation_type.startswith("expect_column_pair"):
            return {
                "domain_kwargs": (
                    "column_A",
                    "column_B",
                    "row_condition",
                    "condition_parser",
                ),
                # NOTE: this is almost certainly incomplete; subclasses should override
                "success_kwargs": (),
                "default_kwarg_values": {
                    "column_A": None,
                    "column_B": None,
                    "row_condition": None,
                    "condition_parser": None,
                },
            }
        elif self.expectation_type.startswith("expect_column"):
            return {
                "domain_kwargs": ("column", "row_condition", "condition_parser"),
                # NOTE: this is almost certainly incomplete; subclasses should override
                "success_kwargs": (),
                "default_kwarg_values": {
                    "column": None,
                    "row_condition": None,
                    "condition_parser": None,
                },
            }

        logger.warning("Requested kwargs for an unrecognized expectation.")
        return {
            "domain_kwargs": (),
            # NOTE: this is almost certainly incomplete; subclasses should override
            "success_kwargs": (),
            "default_kwarg_values": {},
        }

    def get_domain_kwargs(self) -> dict:
        expectation_kwargs_dict: KWargDetailsDict | None = self.kwarg_lookup_dict.get(
            self.expectation_type, None
        )
        if expectation_kwargs_dict is None:
            try:
                impl = get_expectation_impl(self.expectation_type)
            except ExpectationNotFoundError:
                expectation_kwargs_dict = self._get_default_custom_kwargs()
                default_kwarg_values: dict[str, Any] = expectation_kwargs_dict.get(
                    "default_kwarg_values", {}
                )
                domain_keys = expectation_kwargs_dict["domain_kwargs"]
            else:
                domain_keys = impl.domain_keys
                default_kwarg_values = impl.default_kwarg_values
        else:
            default_kwarg_values = expectation_kwargs_dict.get(
                "default_kwarg_values", {}
            )
            domain_keys = expectation_kwargs_dict["domain_kwargs"]

        domain_kwargs = {
            key: self.kwargs.get(key, default_kwarg_values.get(key))
            for key in domain_keys
        }
        missing_kwargs = set(domain_keys) - set(domain_kwargs.keys())
        if missing_kwargs:
            raise InvalidExpectationKwargsError(
                f"Missing domain kwargs: {list(missing_kwargs)}"
            )

        return domain_kwargs

    @public_api
    def get_success_kwargs(self) -> dict:
        """Gets the success and domain kwargs for this ExpectationConfiguration.

        Raises:
            ExpectationNotFoundError: If the expectation implementation is not found.

        Returns:
            A dictionary with the success and domain kwargs of an expectation.
        """
        expectation_kwargs_dict = self.kwarg_lookup_dict.get(
            self.expectation_type, None
        )
        if expectation_kwargs_dict is None:
            try:
                impl = get_expectation_impl(self.expectation_type)
            except ExpectationNotFoundError:
                expectation_kwargs_dict = self._get_default_custom_kwargs()
                default_kwarg_values = expectation_kwargs_dict.get(
                    "default_kwarg_values", {}
                )
                success_keys = expectation_kwargs_dict["success_kwargs"]
            else:
                success_keys = impl.success_keys
                default_kwarg_values = impl.default_kwarg_values
        else:
            default_kwarg_values = expectation_kwargs_dict.get(
                "default_kwarg_values", {}
            )
            success_keys = expectation_kwargs_dict["success_kwargs"]

        domain_kwargs = self.get_domain_kwargs()
        success_kwargs = {
            key: self.kwargs.get(key, default_kwarg_values.get(key))
            for key in success_keys
        }
        success_kwargs.update(domain_kwargs)

        return success_kwargs

    def get_runtime_kwargs(self, runtime_configuration: Optional[dict] = None) -> dict:
        expectation_kwargs_dict = self.kwarg_lookup_dict.get(
            self.expectation_type, None
        )
        runtime_keys: tuple[str, ...]
        if expectation_kwargs_dict is None:
            try:
                impl = get_expectation_impl(self.expectation_type)
            except ExpectationNotFoundError:
                expectation_kwargs_dict = self._get_default_custom_kwargs()
                default_kwarg_values = expectation_kwargs_dict.get(
                    "default_kwarg_values", {}
                )
                runtime_keys = self.runtime_kwargs
            else:
                runtime_keys = impl.runtime_keys
                default_kwarg_values = impl.default_kwarg_values
        else:
            default_kwarg_values = expectation_kwargs_dict.get(
                "default_kwarg_values", {}
            )
            runtime_keys = self.runtime_kwargs

        success_kwargs = self.get_success_kwargs()
        lookup_kwargs = deepcopy(self.kwargs)
        if runtime_configuration:
            lookup_kwargs.update(runtime_configuration)

        runtime_kwargs = {
            key: lookup_kwargs.get(key, default_kwarg_values.get(key))
            for key in runtime_keys
        }
        runtime_kwargs["result_format"] = parse_result_format(
            runtime_kwargs["result_format"]
        )
        runtime_kwargs.update(success_kwargs)

        return runtime_kwargs

    def applies_to_same_domain(
        self, other_expectation_configuration: ExpectationConfiguration
    ) -> bool:
        if (
            not self.expectation_type
            == other_expectation_configuration.expectation_type
        ):
            return False
        return (
            self.get_domain_kwargs()
            == other_expectation_configuration.get_domain_kwargs()
        )

    # noinspection PyPep8Naming
    def isEquivalentTo(
        self,
        other: Union[dict, ExpectationConfiguration],
        match_type: str = "success",
    ) -> bool:
        """ExpectationConfiguration equivalence does not include meta, and relies on *equivalence* of kwargs."""
        if not isinstance(other, self.__class__):
            if isinstance(other, dict):
                try:
                    # noinspection PyNoneFunctionAssignment
                    other = expectationConfigurationSchema.load(other)
                except ValidationError:
                    logger.debug(
                        "Unable to evaluate equivalence of ExpectationConfiguration object with dict because "
                        "dict other could not be instantiated as an ExpectationConfiguration"
                    )
                    return NotImplemented
            else:
                # Delegate comparison to the other instance
                return NotImplemented

        if match_type == "domain":
            return all(
                (
                    self.expectation_type == other.expectation_type,  # type: ignore[union-attr] # could be dict
                    self.get_domain_kwargs() == other.get_domain_kwargs(),  # type: ignore[union-attr] # could be dict
                )
            )

        if match_type == "success":
            return all(
                (
                    self.expectation_type == other.expectation_type,  # type: ignore[union-attr] # could be dict
                    self.get_success_kwargs() == other.get_success_kwargs(),  # type: ignore[union-attr] # could be dict
                )
            )

        if match_type == "runtime":
            return all(
                (
                    self.expectation_type == other.expectation_type,  # type: ignore[union-attr] # could be dict
                    self.kwargs == other.kwargs,  # type: ignore[union-attr] # could be dict
                )
            )

        return False

    def __eq__(self, other):
        """ExpectationConfiguration equality does include meta, but ignores instance identity."""
        if not isinstance(other, self.__class__):
            # Delegate comparison to the other instance's __eq__.
            return NotImplemented
        this_kwargs: dict = convert_to_json_serializable(self.kwargs)
        other_kwargs: dict = convert_to_json_serializable(other.kwargs)
        this_meta: dict = convert_to_json_serializable(self.meta)
        other_meta: dict = convert_to_json_serializable(other.meta)
        return all(
            (
                self.expectation_type == other.expectation_type,
                this_kwargs == other_kwargs,
                this_meta == other_meta,
            )
        )

    def __ne__(self, other):
        # By using the == operator, the returned NotImplemented is handled correctly.
        return not self == other

    def __repr__(self):
        return json.dumps(self.to_json_dict())

    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    @public_api
    def to_json_dict(self) -> Dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this ExpectationConfiguration.

        Returns:
            A JSON-serializable dict representation of this ExpectationConfiguration.
        """
        myself = expectationConfigurationSchema.dump(self)
        # NOTE - JPC - 20191031: migrate to expectation-specific schemas that subclass result with properly-typed
        # schemas to get serialization all-the-way down via dump
        myself["kwargs"] = convert_to_json_serializable(myself["kwargs"])

        # Post dump hook removes this value if null so we need to ensure applicability before conversion
        if "expectation_context" in myself:
            myself["expectation_context"] = convert_to_json_serializable(
                myself["expectation_context"]
            )
        if "rendered_content" in myself:
            myself["rendered_content"] = convert_to_json_serializable(
                myself["rendered_content"]
            )
        return myself

    def get_evaluation_parameter_dependencies(self) -> dict:
        parsed_dependencies: dict = {}
        for value in self.kwargs.values():
            if isinstance(value, dict) and "$PARAMETER" in value:
                param_string_dependencies = find_evaluation_parameter_dependencies(
                    value["$PARAMETER"]
                )
                nested_update(parsed_dependencies, param_string_dependencies)

        dependencies: dict = {}
        urns = parsed_dependencies.get("urns", [])
        for string_urn in urns:
            try:
                urn = ge_urn.parseString(string_urn)
            except ParserError:
                logger.warning("Unable to parse great_expectations urn['$PARAMETER']")
                continue

            # Query stores do not have "expectation_suite_name"
            if urn["urn_type"] == "stores" and "expectation_suite_name" not in urn:
                pass
            else:
                self._update_dependencies_with_expectation_suite_urn(dependencies, urn)

        dependencies = _deduplicate_evaluation_parameter_dependencies(dependencies)

        return dependencies

    @staticmethod
    def _update_dependencies_with_expectation_suite_urn(
        dependencies: dict, urn: ParseResults
    ) -> None:
        if not urn.get("metric_kwargs"):
            nested_update(
                dependencies,
                {urn["expectation_suite_name"]: [urn["metric_name"]]},
            )
        else:
            nested_update(
                dependencies,
                {
                    urn["expectation_suite_name"]: [
                        {
                            "metric_kwargs_id": {
                                urn["metric_kwargs"]: [urn["metric_name"]]
                            }
                        }
                    ]
                },
            )

    def _get_expectation_impl(self) -> Type[Expectation]:
        return get_expectation_impl(self.expectation_type)

    @public_api
    def validate(
        self,
        validator: Validator,
        runtime_configuration: Optional[dict] = None,
    ) -> ExpectationValidationResult:
        """Runs the expectation against a `Validator`.

        Args:
            validator: Object responsible for running an Expectation against data.
            runtime_configuration: A dictionary of configuration arguments to be used by the expectation.

        Raises:
            ExpectationNotFoundError: If the expectation implementation is not found.

        Returns:
            ExpectationValidationResult: The validation result generated by running the expectation against the data.
        """
        expectation_impl: Type[Expectation] = self._get_expectation_impl()
        # noinspection PyCallingNonCallable
        return expectation_impl(self).validate(
            validator=validator,
            runtime_configuration=runtime_configuration,
        )

    def metrics_validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        **kwargs: dict,
    ):
        expectation_impl: Type[Expectation] = self._get_expectation_impl()
        # noinspection PyCallingNonCallable
        return expectation_impl(self).metrics_validate(
            metrics=metrics,
            runtime_configuration=runtime_configuration,
            execution_engine=execution_engine,
        )

    def get_domain_type(self) -> MetricDomainTypes:
        """Return "domain_type" of this expectation."""
        if self.expectation_type.startswith("expect_table_"):
            return MetricDomainTypes.TABLE

        if "column" in self.kwargs:
            return MetricDomainTypes.COLUMN

        if "column_A" in self.kwargs and "column_B" in self.kwargs:
            return MetricDomainTypes.COLUMN_PAIR

        if "column_list" in self.kwargs:
            return MetricDomainTypes.MULTICOLUMN

        raise ValueError(
            'Unable to determine "domain_type" of this "ExpectationConfiguration" object from "kwargs" and heuristics.'
        )

    def render(self) -> None:
        """
        Renders content using the atomic prescriptive renderer for this Expectation Configuration to
            ExpectationConfiguration.rendered_content.
        """
        inline_renderer_config: InlineRendererConfig = {
            "class_name": "InlineRenderer",
            "render_object": self,
        }
        module_name = "great_expectations.render.renderer.inline_renderer"
        inline_renderer = instantiate_class_from_config(
            config=inline_renderer_config,
            runtime_environment={},
            config_defaults={"module_name": module_name},
        )
        if not inline_renderer:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=inline_renderer_config["class_name"],
            )

        self.rendered_content = inline_renderer.get_rendered_content()


class ExpectationConfigurationSchema(Schema):
    expectation_type = fields.Str(
        required=True,
        error_messages={
            "required": "expectation_type missing in expectation configuration"
        },
    )
    kwargs = fields.Dict(
        required=False,
        allow_none=True,
    )
    meta = fields.Dict(
        required=False,
        allow_none=True,
    )
    ge_cloud_id = fields.UUID(required=False, allow_none=True)
    expectation_context = fields.Nested(
        lambda: ExpectationContextSchema, required=False, allow_none=True  # type: ignore[arg-type,return-value]
    )
    rendered_content = fields.List(
        fields.Nested(
            lambda: RenderedAtomicContentSchema, required=False, allow_none=True  # type: ignore[arg-type,return-value]
        )
    )

    REMOVE_KEYS_IF_NONE = ["ge_cloud_id", "expectation_context", "rendered_content"]

    @post_dump
    def clean_null_attrs(self, data: dict, **kwargs: dict) -> dict:
        """Removes the attributes in ExpectationConfigurationSchema.REMOVE_KEYS_IF_NONE during serialization if
        their values are None."""
        data = copy.deepcopy(data)
        for key in ExpectationConfigurationSchema.REMOVE_KEYS_IF_NONE:
            if key in data and data[key] is None:
                data.pop(key)
        return data

    # noinspection PyUnusedLocal
    @post_load
    def make_expectation_configuration(self, data: dict, **kwargs):
        return ExpectationConfiguration(**data)


expectationConfigurationSchema = ExpectationConfigurationSchema()
expectationContextSchema = ExpectationContextSchema()



from typing import TYPE_CHECKING, List, Optional

from great_expectations.core import (
    ExpectationConfiguration,  # noqa: TCH001
    ExpectationValidationResult,  # noqa: TCH001
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    InvalidExpectationConfigurationError,
    render_evaluation_parameter_string,
)
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from great_expectations.rule_based_profiler.config.base import (
    ParameterBuilderConfig,
    RuleBasedProfilerConfig,
)
from great_expectations.rule_based_profiler.parameter_container import (
    DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
    PARAMETER_KEY,
    VARIABLES_KEY,
)

if TYPE_CHECKING:
    from great_expectations.render.renderer_configuration import AddParamArgs


class ExpectColumnValuesToMatchRegex(ColumnMapExpectation):
    """Expect the column entries to be strings that match a given regular expression.

    Valid matches can be found \
    anywhere in the string, for example "[at]+" will identify the following strings as expected: "cat", "hat", \
    "aa", "a", and "t", and the following strings as unexpected: "fish", "dog".

    expect_column_values_to_match_regex is a \
    [Column Map Expectation](https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_column_map_expectations).

    Args:
        column (str): \
            The column name.
        regex (str): \
            The regular expression the column entries should match.

    Keyword Args:
        mostly (None or a float between 0 and 1): \
            Successful if at least mostly fraction of values match the expectation. \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly).

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        include_config (boolean): \
            If True, then include the expectation config as part of the result object.
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, include_config, catch_exceptions, and meta.

    See Also:
        [expect_column_values_to_match_regex_list](https://greatexpectations.io/expectations/expect_column_values_to_match_regex_list)
        [expect_column_values_to_not_match_regex](https://greatexpectations.io/expectations/expect_column_values_to_not_match_regex)
        [expect_column_values_to_not_match_regex_list](https://greatexpectations.io/expectations/expect_column_values_to_not_match_regex_list)
        [expect_column_values_to_match_like_pattern](https://greatexpectations.io/expectations/expect_column_values_to_match_like_pattern)
        [expect_column_values_to_match_like_pattern_list](https://greatexpectations.io/expectations/expect_column_values_to_match_like_pattern_list)
        [expect_column_values_to_not_match_like_pattern](https://greatexpectations.io/expectations/expect_column_values_to_not_match_like_pattern)
        [expect_column_values_to_not_match_like_pattern_list](https://greatexpectations.io/expectations/expect_column_values_to_not_match_like_pattern_list)
    """

    library_metadata = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": [
            "@great_expectations",
        ],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    map_metric = "column_values.match_regex"
    success_keys = (
        "regex",
        "mostly",
        "auto",
        "profiler_config",
    )

    regex_pattern_string_parameter_builder_config: ParameterBuilderConfig = (
        ParameterBuilderConfig(
            module_name="great_expectations.rule_based_profiler.parameter_builder",
            class_name="RegexPatternStringParameterBuilder",
            name="regex_pattern_string_parameter_builder",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=None,
        )
    )
    validation_parameter_builder_configs: List[ParameterBuilderConfig] = [
        regex_pattern_string_parameter_builder_config
    ]
    default_profiler_config = RuleBasedProfilerConfig(
        name="expect_column_values_to_match_regex",  # Convention: use "expectation_type" as profiler name.
        config_version=1.0,
        variables={},
        rules={
            "default_expect_column_values_to_match_regex_rule": {
                "variables": {
                    "mostly": 1.0,
                },
                "domain_builder": {
                    "class_name": "ColumnDomainBuilder",
                    "module_name": "great_expectations.rule_based_profiler.domain_builder",
                },
                "expectation_configuration_builders": [
                    {
                        "expectation_type": "expect_column_values_to_match_regex",
                        "class_name": "DefaultExpectationConfigurationBuilder",
                        "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                        "validation_parameter_builder_configs": validation_parameter_builder_configs,
                        "column": f"{DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}column",
                        "regex": f"{PARAMETER_KEY}{regex_pattern_string_parameter_builder_config.name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY}",
                        "mostly": f"{VARIABLES_KEY}mostly",
                        "meta": {
                            "profiler_details": f"{PARAMETER_KEY}{regex_pattern_string_parameter_builder_config.name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY}",
                        },
                    },
                ],
            },
        },
    )

    default_kwarg_values = {
        "row_condition": None,
        "condition_parser": None,  # we expect this to be explicitly set whenever a row_condition is passed
        "mostly": 1,
        "result_format": "BASIC",
        "include_config": True,
        "catch_exceptions": True,
        "regex": "(?s).*",
        "auto": False,
        "profiler_config": default_profiler_config,
    }
    args_keys = (
        "column",
        "regex",
    )

    def validate_configuration(
        self, configuration: Optional[ExpectationConfiguration] = None
    ) -> None:
        """Validates the configuration for the Expectation.

        For `expect_column_values_to_match_regex`
        we require that the `configuraton.kwargs` contain a `regex` key that is either a `str` or `dict`
        containing `$PARAMETER` key and `str` value.

        Args:
            configuration: The ExpectationConfiguration to be validated.

        Raises:
            InvalidExpectationConfigurationError: The configuraton does not contain the values required by the Expectation
        """
        super().validate_configuration(configuration)
        configuration = configuration or self.configuration

        # supports extensibility by allowing value_set to not be provided in config but captured via child-class default_kwarg_values, e.g. parameterized expectations
        regex = configuration.kwargs.get("regex") or self.default_kwarg_values.get(
            "regex"
        )

        try:
            assert "regex" in configuration.kwargs or regex, "regex is required"
            assert isinstance(regex, (str, dict)), "regex must be a string"
            if isinstance(regex, dict):
                assert (
                    "$PARAMETER" in regex
                ), 'Evaluation Parameter dict for regex kwarg must have "$PARAMETER" key.'
        except AssertionError as e:
            raise InvalidExpectationConfigurationError(str(e))

    @classmethod
    @renderer(renderer_type=LegacyRendererType.QUESTION)
    def _question_renderer(cls, configuration, result=None, runtime_configuration=None):
        column = configuration.kwargs.get("column")
        mostly = configuration.kwargs.get("mostly")
        regex = configuration.kwargs.get("regex")

        return f'Do at least {mostly * 100}% of values in column "{column}" match the regular expression {regex}?'

    @classmethod
    @renderer(renderer_type=LegacyRendererType.ANSWER)
    def _answer_renderer(
        cls, configuration=None, result=None, runtime_configuration=None
    ):
        column = result.expectation_config.kwargs.get("column")
        mostly = result.expectation_config.kwargs.get("mostly")
        regex = result.expectation_config.kwargs.get("regex")
        if result.success:
            return f'At least {mostly * 100}% of values in column "{column}" match the regular expression {regex}.'
        else:
            return f'Less than {mostly * 100}% of values in column "{column}" match the regular expression {regex}.'

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
            ("regex", RendererValueType.STRING),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.regex:
            template_str = (
                "values must match a regular expression but none was specified."
            )
        else:
            template_str = "values must match this regular expression: $regex"
            if params.mostly and params.mostly.value < 1.0:
                renderer_configuration = cls._add_mostly_pct_param(
                    renderer_configuration=renderer_configuration
                )
                template_str += ", at least $mostly_pct % of the time."
            else:
                template_str += "."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_evaluation_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = (
            False if runtime_configuration.get("include_column_name") is False else True
        )
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            ["column", "regex", "mostly", "row_condition", "condition_parser"],
        )

        if not params.get("regex"):
            template_str = (
                "values must match a regular expression but none was specified."
            )
        else:
            template_str = "values must match this regular expression: $regex"
            if params["mostly"] is not None and params["mostly"] < 1.0:
                params["mostly_pct"] = num_to_str(
                    params["mostly"] * 100, precision=15, no_scientific=True
                )
                # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")
                template_str += ", at least $mostly_pct % of the time."
            else:
                template_str += "."

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        params_with_json_schema = {  # noqa: F841 # unused
            "column": {"schema": {"type": "string"}, "value": params.get("column")},
            "mostly": {"schema": {"type": "number"}, "value": params.get("mostly")},
            "mostly_pct": {
                "schema": {"type": "number"},
                "value": params.get("mostly_pct"),
            },
            "regex": {"schema": {"type": "string"}, "value": params.get("regex")},
            "row_condition": {
                "schema": {"type": "string"},
                "value": params.get("row_condition"),
            },
            "condition_parser": {
                "schema": {"type": "string"},
                "value": params.get("condition_parser"),
            },
        }

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    # examples = [
    #     {
    #         "data": {
    #             "a": ["aaa", "abb", "acc", "add", "bee"],
    #             "b": ["aaa", "abb", "acc", "bdd", None],
    #             "column_name with space": ["aaa", "abb", "acc", "add", "bee"],
    #         },
    #         "tests": [
    #             {
    #                 "title": "negative_test_insufficient_mostly_and_one_non_matching_value",
    #                 "exact_match_out": False,
    #                 "in": {"column": "a", "regex": "^a", "mostly": 0.9},
    #                 "out": {
    #                     "success": False,
    #                     "unexpected_index_list": [4],
    #                     "unexpected_list": ["bee"],
    #                 },
    #                 "include_in_gallery": True,
    #                 "suppress_test_for": ["sqlite", "mssql"],
    #             },
    #             {
    #                 "title": "positive_test_exact_mostly_w_one_non_matching_value",
    #                 "exact_match_out": False,
    #                 "in": {"column": "a", "regex": "^a", "mostly": 0.8},
    #                 "out": {
    #                     "success": True,
    #                     "unexpected_index_list": [4],
    #                     "unexpected_list": ["bee"],
    #                 },
    #                 "include_in_gallery": True,
    #                 "suppress_test_for": ["sqlite", "mssql"],
    #             },
    #         ],
    #     }
    # ]



from __future__ import annotations

import copy
import logging
import pathlib
from typing import TYPE_CHECKING, ClassVar, Optional, Set

import great_expectations.exceptions as ge_exceptions
from great_expectations.core.batch_spec import PathBatchSpec
from great_expectations.experimental.datasources.data_asset.data_connector import (
    DataConnector,
    FilesystemDataConnector,
)
from great_expectations.experimental.datasources.file_path_data_asset import (
    _FilePathDataAsset,
)
from great_expectations.experimental.datasources.interfaces import (
    Batch,
    BatchRequest,
    TestConnectionError,
)

if TYPE_CHECKING:
    from great_expectations.execution_engine import (
        PandasExecutionEngine,
        SparkDFExecutionEngine,
    )
    from great_expectations.experimental.datasources.interfaces import (
        BatchRequestOptions,
    )

logger = logging.getLogger(__name__)


class _FilesystemDataAsset(_FilePathDataAsset):
    _EXCLUDE_FROM_READER_OPTIONS: ClassVar[
        Set[str]
    ] = _FilePathDataAsset._EXCLUDE_FROM_READER_OPTIONS | {
        "glob_directive",
    }

    # Filesystem specific attributes
    glob_directive: str = "**/*"

    def _get_reader_options_include(self) -> set[str] | None:
        raise NotImplementedError(
            """One needs to explicitly provide set(str)-valued reader options for "pydantic.BaseModel.dict()" method \
to use as its "include" directive for Filesystem style DataAsset processing."""
        )

    def test_connection(self) -> None:
        """Test the connection for the FilesystemDataAsset.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        success = False
        for filepath in self.datasource.base_directory.glob("**/*.*"):
            if self.batching_regex.match(
                str(filepath.relative_to(self.datasource.base_directory))
            ):
                # if one file in the path matches the batching_regex, we consider this asset valid
                success = True
                break
        if not success:
            raise TestConnectionError(self._build_test_connection_error_message())

    def _fully_specified_batch_requests_with_path(
        self, batch_request: BatchRequest
    ) -> list[tuple[BatchRequest, pathlib.Path]]:
        """Generates a list fully specified batch requests from partial specified batch request

        Args:
            batch_request: A batch request

        Returns:
            A list of pairs (batch_request, path) where 'batch_request' is a fully specified
            batch request and 'path' is the path to the corresponding file on disk.
            This list will be empty if no files exist on disk that correspond to the input
            batch request.
        """
        base_directory: pathlib.Path = self.datasource.base_directory
        all_paths: list[pathlib.Path] = [
            pathlib.Path(path) for path in base_directory.glob("**/*.*")
        ]

        batch_requests_with_path: list[tuple[BatchRequest, pathlib.Path]] = []

        abs_path: pathlib.Path
        for abs_path in all_paths:
            path_relative_to_base_dir = abs_path.relative_to(base_directory)
            match = self._regex_parser.get_matches(
                target=str(path_relative_to_base_dir)
            )
            if match:
                # Create the batch request that would correlate to this regex match
                match_options = {}
                for group_id in range(
                    1, self._regex_parser.get_num_all_matched_group_values() + 1
                ):
                    match_options[
                        self._all_group_index_to_group_name_mapping[group_id]
                    ] = match.group(group_id)
                    match_options["path"] = path_relative_to_base_dir
                # Determine if this file_name matches the batch_request
                allowed_match = True
                for key, value in batch_request.options.items():
                    if key == "path" and isinstance(value, str):
                        value = pathlib.Path(value)
                    if match_options[key] != value:
                        allowed_match = False
                        break
                if allowed_match:
                    batch_requests_with_path.append(
                        (
                            BatchRequest(
                                datasource_name=self.datasource.name,
                                data_asset_name=self.name,
                                options=match_options,
                            ),
                            abs_path,
                        )
                    )
                    logger.debug(f"Matching path: {abs_path}")
        if not batch_requests_with_path:
            logger.warning(
                f"Batch request {batch_request} corresponds to no data files."
            )
        return batch_requests_with_path

    def batch_request_options_template(
        self,
    ) -> BatchRequestOptions:
        options: set[str] = copy.deepcopy(
            self._ALWAYS_INCLUDE_IN_BATCH_REQUEST_OPTIONS_TEMPLATE
        )
        options.update(set(self._all_group_names))
        return {option: None for option in options}

    def build_batch_request(
        self, options: Optional[BatchRequestOptions] = None
    ) -> BatchRequest:
        if options:
            for option, value in options.items():
                if (
                    option in self._all_group_name_to_group_index_mapping
                    and not isinstance(value, str)
                ):
                    raise ge_exceptions.InvalidBatchRequestError(
                        f"All batching_regex matching options must be strings. The value of '{option}' is "
                        f"not a string: {value}"
                    )

        return super().build_batch_request(options)

    def get_batch_list_from_batch_request(
        self, batch_request: BatchRequest
    ) -> list[Batch]:
        self._validate_batch_request(batch_request)
        batch_list: list[Batch] = []

        kwargs: dict | None = getattr(self, "kwargs", None)
        if not kwargs:
            kwargs = {}

        for request, path in self._fully_specified_batch_requests_with_path(
            batch_request
        ):
            batch_spec = PathBatchSpec(
                path=str(path),
                reader_method=self._get_reader_method(),
                reader_options=self.dict(
                    include=self._get_reader_options_include(),
                    exclude=self._EXCLUDE_FROM_READER_OPTIONS,
                    exclude_unset=True,
                    by_alias=True,
                    **kwargs,
                ),
            )
            execution_engine: PandasExecutionEngine | SparkDFExecutionEngine = (
                self.datasource.get_execution_engine()
            )
            data, markers = execution_engine.get_batch_data_and_markers(
                batch_spec=batch_spec
            )

            # batch_definition (along with batch_spec and markers) is only here to satisfy a
            # legacy constraint when computing usage statistics in a validator. We hope to remove
            # it in the future.
            # imports are done inline to prevent a circular dependency with core/batch.py
            from great_expectations.core import IDDict
            from great_expectations.core.batch import BatchDefinition

            batch_definition = BatchDefinition(
                datasource_name=self.datasource.name,
                data_connector_name="experimental",
                data_asset_name=self.name,
                batch_identifiers=IDDict(request.options),
                batch_spec_passthrough=None,
            )

            batch_metadata = copy.deepcopy(request.options)
            batch_metadata["base_directory"] = self.datasource.base_directory

            # Some pydantic annotations are postponed due to circular imports.
            # Batch.update_forward_refs() will set the annotations before we
            # instantiate the Batch class since we can import them in this scope.
            Batch.update_forward_refs()
            batch_list.append(
                Batch(
                    datasource=self.datasource,
                    data_asset=self,
                    batch_request=request,
                    data=data,
                    metadata=batch_metadata,
                    legacy_batch_markers=markers,
                    legacy_batch_spec=batch_spec,
                    legacy_batch_definition=batch_definition,
                )
            )
        self.sort_batches(batch_list)
        return batch_list

    def _build_test_connection_error_message(self) -> str:
        return f"""No file at base_directory path "{self.datasource.base_directory.resolve()}" matched regular expressions pattern "{self.batching_regex.pattern}" and/or glob_directive "{self.glob_directive}" for DataAsset "{self.name}"."""

    def _build_data_connector(self) -> DataConnector:
        data_connector: DataConnector = FilesystemDataConnector(
            datasource_name=self.datasource.name,
            data_asset_name=self.name,
            base_directory=self.datasource.base_directory,
            batching_regex=self.batching_regex,
            glob_directive=self.glob_directive,
            data_context_root_directory=self.datasource.data_context_root_directory,
        )
        return data_connector



# isort:skip_file

from .data_connector import DataConnector
from .runtime_data_connector import RuntimeDataConnector
from .file_path_data_connector import FilePathDataConnector
from .configured_asset_file_path_data_connector import (
    ConfiguredAssetFilePathDataConnector,
)
from .inferred_asset_file_path_data_connector import (
    InferredAssetFilePathDataConnector,
)
from .configured_asset_filesystem_data_connector import (
    ConfiguredAssetFilesystemDataConnector,
)
from .inferred_asset_filesystem_data_connector import (
    InferredAssetFilesystemDataConnector,
)
from .configured_asset_s3_data_connector import (
    ConfiguredAssetS3DataConnector,
)
from .inferred_asset_s3_data_connector import InferredAssetS3DataConnector
from .configured_asset_azure_data_connector import (
    ConfiguredAssetAzureDataConnector,
)
from .inferred_asset_azure_data_connector import (
    InferredAssetAzureDataConnector,
)
from .configured_asset_gcs_data_connector import (
    ConfiguredAssetGCSDataConnector,
)
from .inferred_asset_gcs_data_connector import (
    InferredAssetGCSDataConnector,
)
from .configured_asset_sql_data_connector import (
    ConfiguredAssetSqlDataConnector,
)
from .inferred_asset_sql_data_connector import (
    InferredAssetSqlDataConnector,
)
from .configured_asset_dbfs_data_connector import ConfiguredAssetDBFSDataConnector
from .inferred_asset_dbfs_data_connector import InferredAssetDBFSDataConnector
from .configured_asset_aws_glue_data_catalog_data_connector import (
    ConfiguredAssetAWSGlueDataCatalogDataConnector,
)
from .inferred_asset_aws_glue_data_catalog_data_connector import (
    InferredAssetAWSGlueDataCatalogDataConnector,
)



from typing import Dict, List, Optional, Union

from great_expectations.core._docs_decorators import public_api
from great_expectations.datasource.data_connector.configured_asset_sql_data_connector import (
    ConfiguredAssetSqlDataConnector,
)
from great_expectations.execution_engine import ExecutionEngine  # noqa: TCH001
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.util import deep_filter_properties_iterable

try:
    import sqlalchemy as sa
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.reflection import Inspector
    from sqlalchemy.exc import OperationalError
except ImportError:
    sa = None
    Engine = None
    Inspector = None
    OperationalError = None


@public_api
class InferredAssetSqlDataConnector(ConfiguredAssetSqlDataConnector):
    """An Inferred Asset Data Connector used to connect to an SQL database.

    This Data Connector determines Data Asset names by introspecting the database schema.

    Args:
        name: The name of the Data Connector.
        datasource_name: The name of this Data Connector's Datasource.
        execution_engine: The Execution Engine object to used by this Data Connector to read the data.
        data_asset_name_prefix: A prefix to prepend to all names of Data Assets inferred by this Data Connector.
        data_asset_name_suffix: A suffix to append to all names of Data Asset inferred by this Data Connector.
        include_schema_name: If True the Data Asset name  will include the schema as a prefix.
        splitter_method: A method to use to split the target table into multiple Batches.
        splitter_kwargs: Keyword arguments to pass to the splitter method.
        sampling_method: A method to use to downsample within a target Batch.
        sampling_kwargs: Keyword arguments to pass to sampling method.
        excluded_tables: A list of tables to ignore when inferring Data Asset names.
        included_tables: A list of tables to include when inferring Data Asset names. When provided, only Data Assets
            matching this list will be inferred.
        skip_inapplicable_tables: If True, tables that can't be successfully queried using sampling and splitter methods
            are excluded from inferred data_asset_names. If False, the class will throw an error during initialization
            if any such tables are encountered.
        introspection_directives: Arguments passed to the introspection method to guide introspection. These may be,
            `schema_name`, which filters to a specific schema,
            `ignore_information_schemas_and_system_tables`, which defaults to True,
            `information_schemas`, a list of schemas to consider as informational,
            `system_tables`, a list of tables to consider system tables,
            and `include_views` which defaults to True.
        batch_spec_passthrough: Dictionary with keys that will be added directly to the batch spec.
        id: The unique identifier for this Data Connector used when running in cloud mode.
    """

    def __init__(
        self,
        name: str,
        datasource_name: str,
        execution_engine: Optional[ExecutionEngine] = None,
        data_asset_name_prefix: str = "",
        data_asset_name_suffix: str = "",
        include_schema_name: bool = False,
        splitter_method: Optional[str] = None,
        splitter_kwargs: Optional[dict] = None,
        sampling_method: Optional[str] = None,
        sampling_kwargs: Optional[dict] = None,
        excluded_tables: Optional[list] = None,
        included_tables: Optional[list] = None,
        skip_inapplicable_tables: bool = True,
        introspection_directives: Optional[dict] = None,
        batch_spec_passthrough: Optional[dict] = None,
        id: Optional[str] = None,
    ) -> None:
        super().__init__(
            name=name,
            datasource_name=datasource_name,
            execution_engine=execution_engine,
            include_schema_name=include_schema_name,
            splitter_method=splitter_method,
            splitter_kwargs=splitter_kwargs,
            sampling_method=sampling_method,
            sampling_kwargs=sampling_kwargs,
            assets=None,
            batch_spec_passthrough=batch_spec_passthrough,
            id=id,
        )

        self._data_asset_name_prefix = data_asset_name_prefix
        self._data_asset_name_suffix = data_asset_name_suffix

        self._excluded_tables = excluded_tables
        self._included_tables = included_tables
        self._skip_inapplicable_tables = skip_inapplicable_tables

        if introspection_directives is None:
            introspection_directives = {}

        self._introspection_directives = introspection_directives
        # This cache will contain a "config" for each data_asset discovered via introspection.
        # This approach ensures that ConfiguredAssetSqlDataConnector._assets and _introspected_assets_cache store objects of the same "type"
        # Note: We should probably turn them into AssetConfig objects

        self._refresh_introspected_assets_cache()

    @property
    def data_asset_name_prefix(self) -> str:
        return self._data_asset_name_prefix

    @property
    def data_asset_name_suffix(self) -> str:
        return self._data_asset_name_suffix

    def _refresh_data_references_cache(self) -> None:
        self._refresh_introspected_assets_cache()
        super()._refresh_data_references_cache()

    def _refresh_introspected_assets_cache(self) -> None:
        introspected_table_metadata = self._introspect_db(
            **self._introspection_directives
        )

        introspected_assets: dict = {}

        for metadata in introspected_table_metadata:
            if (self._excluded_tables is not None) and (
                f"{metadata['schema_name']}.{metadata['table_name']}"
                in self._excluded_tables
            ):
                continue

            if (self._included_tables is not None) and (
                f"{metadata['schema_name']}.{metadata['table_name']}"
                not in self._included_tables
            ):
                continue

            schema_name: str = metadata["schema_name"]
            table_name: str = metadata["table_name"]

            data_asset_config: dict = deep_filter_properties_iterable(
                properties={
                    "type": metadata["type"],
                    "table_name": table_name,
                    "data_asset_name_prefix": self.data_asset_name_prefix,
                    "data_asset_name_suffix": self.data_asset_name_suffix,
                    "include_schema_name": self.include_schema_name,
                    "schema_name": schema_name,
                    "splitter_method": self.splitter_method,
                    "splitter_kwargs": self.splitter_kwargs,
                    "sampling_method": self.sampling_method,
                    "sampling_kwargs": self.sampling_kwargs,
                },
            )

            data_asset_name: str = self._update_data_asset_name_from_config(
                data_asset_name=table_name, data_asset_config=data_asset_config
            )

            # Attempt to fetch a list of batch_identifiers from the table
            try:
                self._get_batch_identifiers_list_from_data_asset_config(
                    data_asset_name=data_asset_name,
                    data_asset_config=data_asset_config,
                )
            except OperationalError as e:
                # If it doesn't work, then...
                if self._skip_inapplicable_tables:
                    # No harm done. Just don't include this table in the list of assets.
                    continue
                else:
                    # We're being strict. Crash now.
                    raise ValueError(
                        f"Couldn't execute a query against table {metadata['table_name']} in schema {metadata['schema_name']}"
                    ) from e

            # Store an asset config for each introspected data asset.
            introspected_assets[data_asset_name] = data_asset_config
            self.add_data_asset(name=table_name, config=data_asset_config)

    def _introspect_db(  # noqa: C901 - 16
        self,
        schema_name: Union[str, None] = None,
        ignore_information_schemas_and_system_tables: bool = True,
        information_schemas: Optional[List[str]] = None,
        system_tables: Optional[List[str]] = None,
        include_views=True,
    ):
        if information_schemas is None:
            information_schemas = [
                "INFORMATION_SCHEMA",  # snowflake, mssql, mysql, oracle
                "information_schema",  # postgres, redshift, mysql
                "performance_schema",  # mysql
                "sys",  # mysql
                "mysql",  # mysql
            ]

        if system_tables is None:
            system_tables = ["sqlite_master"]  # sqlite

        engine: Engine = self.execution_engine.engine
        inspector: Inspector = sa.inspect(engine)

        selected_schema_name = schema_name

        tables: List[Dict[str, str]] = []
        schema_names: List[str] = inspector.get_schema_names()
        for schema_name in schema_names:
            if (
                ignore_information_schemas_and_system_tables
                and schema_name in information_schemas
            ):
                continue

            if selected_schema_name is not None and schema_name != selected_schema_name:
                continue

            table_names: List[str] = inspector.get_table_names(schema=schema_name)
            for table_name in table_names:
                if ignore_information_schemas_and_system_tables and (
                    table_name in system_tables
                ):
                    continue

                tables.append(
                    {
                        "schema_name": schema_name,
                        "table_name": table_name,
                        "type": "table",
                    }
                )

            # Note Abe 20201112: This logic is currently untested.
            if include_views:
                # Note: this is not implemented for bigquery
                try:
                    view_names = inspector.get_view_names(schema=schema_name)
                except NotImplementedError:
                    # Not implemented by Athena dialect
                    pass
                else:
                    for view_name in view_names:
                        if ignore_information_schemas_and_system_tables and (
                            view_name in system_tables
                        ):
                            continue

                        tables.append(
                            {
                                "schema_name": schema_name,
                                "table_name": view_name,
                                "type": "view",
                            }
                        )

        # SQLAlchemy's introspection does not list "external tables" in Redshift Spectrum (tables whose data is stored on S3).
        # The following code fetches the names of external schemas and tables from a special table
        # 'svv_external_tables'.
        try:
            if engine.dialect.name.lower() == GXSqlDialect.REDSHIFT:
                # noinspection SqlDialectInspection,SqlNoDataSourceInspection
                result = engine.execute(
                    "select schemaname, tablename from svv_external_tables"
                ).fetchall()
                for row in result:
                    tables.append(
                        {
                            "schema_name": row[0],
                            "table_name": row[1],
                            "type": "table",
                        }
                    )
        except Exception as e:
            # Our testing shows that 'svv_external_tables' table is present in all Redshift clusters. This means that this
            # exception is highly unlikely to fire.
            if "UndefinedTable" not in str(e):
                raise e

        return tables



from typing import Union

import pytest

from great_expectations.exceptions import ProfilerConfigurationError
from great_expectations.rule_based_profiler.domain_builder.categorical_column_domain_builder import (
    AbsoluteCardinalityLimit,
    CardinalityChecker,
    CardinalityLimitMode,
    RelativeCardinalityLimit,
)


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_cardinality_limit_mode_parameter_absolute():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        cardinality_limit_mode=CardinalityLimitMode.MANY
    )
    assert cardinality_checker.cardinality_limit_mode == CardinalityLimitMode.MANY.value
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, AbsoluteCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == AbsoluteCardinalityLimit(
        name="MANY", max_unique_values=10000
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_cardinality_limit_mode_parameter_relative():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        cardinality_limit_mode=CardinalityLimitMode.REL_1
    )
    assert (
        cardinality_checker.cardinality_limit_mode == CardinalityLimitMode.REL_1.value
    )
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, RelativeCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == RelativeCardinalityLimit(
        name="REL_1", max_proportion_unique=0.01
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_cardinality_limit_mode_parameter_str_absolute():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        cardinality_limit_mode="VERY_MANY"
    )
    assert (
        cardinality_checker.cardinality_limit_mode
        == CardinalityLimitMode.VERY_MANY.value
    )
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, AbsoluteCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == AbsoluteCardinalityLimit(
        name="VERY_MANY", max_unique_values=100000
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_cardinality_limit_mode_parameter_str_relative():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        cardinality_limit_mode="REL_0_1"
    )
    assert (
        cardinality_checker.cardinality_limit_mode == CardinalityLimitMode.REL_0_1.value
    )
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, RelativeCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == RelativeCardinalityLimit(
        name="REL_0_1", max_proportion_unique=0.001
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_max_unique_values_parameter():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        max_unique_values=12345
    )
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, AbsoluteCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == AbsoluteCardinalityLimit(
        name="CUSTOM_ABS_12345", max_unique_values=12345
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_max_proportion_unique_parameter_float():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        max_proportion_unique=0.42
    )
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, RelativeCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == RelativeCardinalityLimit(
        name="CUSTOM_REL_0.42", max_proportion_unique=0.42
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_valid_max_proportion_unique_parameter_int():
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        max_proportion_unique=1
    )
    assert isinstance(
        cardinality_checker.cardinality_limit_mode, RelativeCardinalityLimit
    )
    assert cardinality_checker.cardinality_limit_mode == RelativeCardinalityLimit(
        name="CUSTOM_REL_1", max_proportion_unique=1
    )


@pytest.mark.unit
def test_cardinality_checker_instantiation_invalid_cardinality_limit_mode_parameter():
    with pytest.raises(ProfilerConfigurationError) as excinfo:
        # noinspection PyUnusedLocal,PyTypeChecker
        cardinality_checker: CardinalityChecker = CardinalityChecker(
            cardinality_limit_mode="&^*%not_a_valid_mode^&*("
        )
    assert "specify a supported cardinality mode" in str(excinfo.value)
    assert "REL_1" in str(excinfo.value)
    assert "MANY" in str(excinfo.value)


@pytest.mark.unit
def test_cardinality_checker_instantiation_invalid_max_unique_values_parameter():
    with pytest.raises(ProfilerConfigurationError) as excinfo:
        # noinspection PyUnusedLocal,PyTypeChecker
        cardinality_checker: CardinalityChecker = CardinalityChecker(
            max_unique_values="invalid_input_type_string"
        )
    assert "specify an int" in str(excinfo.value)
    assert "str" in str(excinfo.value)


@pytest.mark.unit
def test_cardinality_checker_instantiation_invalid_max_proportion_unique_parameter():
    with pytest.raises(ProfilerConfigurationError) as excinfo:
        # noinspection PyUnusedLocal,PyTypeChecker
        cardinality_checker: CardinalityChecker = CardinalityChecker(
            max_proportion_unique="invalid_input_type_string"
        )
    assert "specify a float or int" in str(excinfo.value)
    assert "str" in str(excinfo.value)


@pytest.mark.unit
def test_cardinality_checker_instantiation_invalid_multiple_parameters():
    with pytest.raises(ProfilerConfigurationError) as excinfo:
        # noinspection PyUnusedLocal,PyTypeChecker
        cardinality_checker: CardinalityChecker = CardinalityChecker(
            cardinality_limit_mode="REL_0_1", max_proportion_unique=0.42
        )
    assert "Please pass ONE of the following parameters" in str(excinfo.value)
    assert "you passed 2 parameters" in str(excinfo.value)


@pytest.mark.parametrize(
    "input_cardinality_limit_mode",
    [member for member in CardinalityLimitMode]
    + [member.name for member in CardinalityLimitMode],
)
@pytest.mark.unit
def test_exhaustively_cardinality_within_limit_for_all_supported_cardinality_limits(
    input_cardinality_limit_mode,
):
    """What does this test and why?

    This test checks that the CardinalityChecker appropriately handles all types of limits,
    whether created using a CardinalityLimitMode or a string.
    """

    cardinality_checker: CardinalityChecker = CardinalityChecker(
        cardinality_limit_mode=input_cardinality_limit_mode
    )

    # Set up passing and failing cardinality measured values
    cardinality_limit: float
    cardinality_limit_mode: Union[
        AbsoluteCardinalityLimit, RelativeCardinalityLimit
    ] = cardinality_checker.cardinality_limit_mode
    if isinstance(cardinality_limit_mode, AbsoluteCardinalityLimit):
        cardinality_limit = cardinality_limit_mode.max_unique_values
        if cardinality_limit > 0:
            passing_cardinality = cardinality_limit - 1
        else:
            passing_cardinality = 0
        failing_cardinality = cardinality_limit + 1
    elif isinstance(cardinality_limit_mode, RelativeCardinalityLimit):
        cardinality_limit = cardinality_limit_mode.max_proportion_unique
        passing_cardinality = 0.0
        failing_cardinality = cardinality_limit + 0.01
    else:
        raise ValueError("cardinality_limit_mode should be a supported type")

    assert cardinality_checker.cardinality_within_limit(passing_cardinality)
    assert not cardinality_checker.cardinality_within_limit(failing_cardinality)


@pytest.mark.unit
def test_cardinality_checker_cardinality_within_limit_max_unique_values_parameter():
    input_max_unique_values: int = 12345
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        max_unique_values=input_max_unique_values
    )
    passing_cardinality: int = input_max_unique_values - 1
    failing_cardinality: int = input_max_unique_values + 1
    assert cardinality_checker.cardinality_within_limit(passing_cardinality)
    assert not cardinality_checker.cardinality_within_limit(failing_cardinality)


@pytest.mark.unit
def test_cardinality_checker_cardinality_within_limit_max_proportion_unique_parameter_float():
    input_max_proportion_unique: float = 0.42
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        max_proportion_unique=input_max_proportion_unique
    )
    passing_cardinality: float = input_max_proportion_unique - 0.01
    failing_cardinality: float = input_max_proportion_unique + 0.01
    assert cardinality_checker.cardinality_within_limit(passing_cardinality)
    assert not cardinality_checker.cardinality_within_limit(failing_cardinality)


@pytest.mark.unit
def test_cardinality_checker_cardinality_within_limit_max_proportion_unique_parameter_int():
    input_max_proportion_unique: int = 1
    cardinality_checker: CardinalityChecker = CardinalityChecker(
        max_proportion_unique=input_max_proportion_unique
    )
    passing_cardinality: float = input_max_proportion_unique - 0.01
    failing_cardinality: float = input_max_proportion_unique + 0.01
    assert cardinality_checker.cardinality_within_limit(passing_cardinality)
    assert not cardinality_checker.cardinality_within_limit(failing_cardinality)



from typing import TYPE_CHECKING, Dict, Optional

import numpy as np
import pandas as pd

from great_expectations.core import (
    ExpectationConfiguration,  # noqa: TCH001
    ExpectationValidationResult,  # noqa: TCH001
)
from great_expectations.execution_engine import PandasExecutionEngine
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    InvalidExpectationConfigurationError,
    render_evaluation_parameter_string,
)
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.render.renderer_configuration import AddParamArgs


class ExpectColumnValuesToNotBeInSet(ColumnMapExpectation):
    """Expect column entries to not be in the set.

    For example:
    ::

        # my_df.my_col = [1,2,2,3,3,3]
        >>> my_df.expect_column_values_to_not_be_in_set(
                "my_col",
                [1,2]
            )
        {
            "success": false
            "result": {
                "unexpected_count": 3
                "unexpected_percent": 50.0,
                "unexpected_percent_nonmissing": 50.0,
                "partial_unexpected_list": [
                    1, 2, 2
                ],
            },
        }

    expect_column_values_to_not_be_in_set is a \
    [Column Map Expectation](https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_column_map_expectations).

    Args:
        column (str): \
            The column name.
        value_set (set-like): \
            A set of objects used for comparison.

    Keyword Args:
        mostly (None or a float between 0 and 1): \
            Successful if at least mostly fraction of values match the expectation. \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly).

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        include_config (boolean): \
            If True, then include the expectation config as part of the result object.
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, include_config, catch_exceptions, and meta.

    See Also:
        [expect_column_values_to_be_in_set](https://greatexpectations.io/expectations/expect_column_values_to_be_in_set)
    """

    library_metadata = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": [
            "@great_expectations",
        ],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    map_metric = "column_values.not_in_set"
    success_keys = (
        "value_set",
        "mostly",
        "parse_strings_as_datetimes",
    )
    default_kwarg_values = {
        "row_condition": None,
        "condition_parser": None,  # we expect this to be explicitly set whenever a row_condition is passed
        "mostly": 1,
        "parse_strings_as_datetimes": False,
        "result_format": "BASIC",
        "include_config": True,
        "catch_exceptions": False,
    }
    args_keys = (
        "column",
        "value_set",
    )

    def validate_configuration(
        self, configuration: Optional[ExpectationConfiguration] = None
    ) -> None:
        """
        Validates the configuration of an Expectation.

        The configuration will also be validated using each of the `validate_configuration` methods in its Expectation
        superclass hierarchy.

        Args:
            configuration: An `ExpectationConfiguration` to validate. If no configuration is provided, it will be pulled
                                  from the configuration attribute of the Expectation instance.

        Raises:
            `InvalidExpectationConfigurationError`: The configuration does not contain the values required by the
            Expectation.
        """
        super().validate_configuration(configuration)
        configuration = configuration or self.configuration
        try:
            assert "value_set" in configuration.kwargs, "value_set is required"
            assert isinstance(
                configuration.kwargs["value_set"], (list, set, dict)
            ), "value_set must be a list or a set"
            if isinstance(configuration.kwargs["value_set"], dict):
                assert (
                    "$PARAMETER" in configuration.kwargs["value_set"]
                ), 'Evaluation Parameter dict for value_set kwarg must have "$PARAMETER" key.'
        except AssertionError as e:
            raise InvalidExpectationConfigurationError(str(e))

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("value_set", RendererValueType.ARRAY),
            ("mostly", RendererValueType.NUMBER),
            ("parse_strings_as_datetimes", RendererValueType.BOOLEAN),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params
        template_str = ""

        if params.value_set:
            array_param_name = "value_set"
            param_prefix = "v__"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            value_set_str: str = cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            template_str += f"values must not belong to this set: {value_set_str}"

            if params.mostly and params.mostly.value < 1.0:
                renderer_configuration = cls._add_mostly_pct_param(
                    renderer_configuration=renderer_configuration
                )
                template_str += ", at least $mostly_pct % of the time."
            else:
                template_str += "."

            if params.parse_strings_as_datetimes:
                template_str += " Values should be parsed as datetimes."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_evaluation_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        renderer_configuration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        params = substitute_none_for_missing(
            renderer_configuration.kwargs,
            [
                "column",
                "value_set",
                "mostly",
                "parse_strings_as_datetimes",
                "row_condition",
                "condition_parser",
            ],
        )

        if params["value_set"] is None or len(params["value_set"]) == 0:
            values_string = "[ ]"
        else:
            for i, v in enumerate(params["value_set"]):
                params[f"v__{str(i)}"] = v

            values_string = " ".join(
                [f"$v__{str(i)}" for i, v in enumerate(params["value_set"])]
            )

        template_str = f"values must not belong to this set: {values_string}"

        if params["mostly"] is not None and params["mostly"] < 1.0:
            params["mostly_pct"] = num_to_str(
                params["mostly"] * 100, precision=15, no_scientific=True
            )
            # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")
            template_str += ", at least $mostly_pct % of the time."
        else:
            template_str += "."

        if params.get("parse_strings_as_datetimes"):
            template_str += " Values should be parsed as datetimes."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        styling = (
            runtime_configuration.get("styling", {}) if runtime_configuration else {}
        )

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    def _pandas_column_values_not_in_set(
        self,
        series: pd.Series,
        metrics: Dict,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        runtime_configuration: Optional[dict] = None,
        filter_column_isnull: bool = True,
    ):
        value_set = metric_value_kwargs["value_set"]

        if value_set is None:
            # Vacuously true
            return np.ones(len(series), dtype=np.bool_)
        if pd.api.types.is_datetime64_any_dtype(series):
            parsed_value_set = PandasExecutionEngine.parse_value_set(
                value_set=value_set
            )
        else:
            parsed_value_set = value_set

        return pd.DataFrame(
            {"column_values.not_in_set": ~series.isin(parsed_value_set)}
        )



from typing import List

from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import Batch, BatchRequest, RuntimeBatchRequest

context = gx.get_context()

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/s3/pandas/inferred_and_runtime_python_example.py datasource_config">
datasource_config = {
    "name": "my_s3_datasource",
    "class_name": "Datasource",
    "execution_engine": {"class_name": "PandasExecutionEngine"},
    "data_connectors": {
        "default_runtime_data_connector_name": {
            "class_name": "RuntimeDataConnector",
            "batch_identifiers": ["default_identifier_name"],
        },
        "default_inferred_data_connector_name": {
            "class_name": "InferredAssetS3DataConnector",
            "bucket": "<YOUR_S3_BUCKET_HERE>",
            "prefix": "<BUCKET_PATH_TO_DATA>",
            "default_regex": {
                "pattern": "(.*)\\.csv",
                "group_names": ["data_asset_name"],
            },
        },
    },
}
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_config["data_connectors"]["default_inferred_data_connector_name"][
    "bucket"
] = "superconductive-docs-test"
datasource_config["data_connectors"]["default_inferred_data_connector_name"][
    "prefix"
] = "data/taxi_yellow_tripdata_samples/"

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/s3/pandas/inferred_and_runtime_python_example.py test_yaml_config">
context.test_yaml_config(yaml.dump(datasource_config))
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/s3/pandas/inferred_and_runtime_python_example.py add_datasource">
context.add_datasource(**datasource_config)
# </snippet>

# Here is a RuntimeBatchRequest using a path to a single CSV file
batch_request = RuntimeBatchRequest(
    datasource_name="my_s3_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="<YOUR_MEANGINGFUL_NAME>",  # this can be anything that identifies this data_asset for you
    runtime_parameters={"path": "<PATH_TO_YOUR_DATA_HERE>"},  # Add your S3 path here.
    batch_identifiers={"default_identifier_name": "default_identifier"},
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the BatchRequest above.
batch_request.runtime_parameters[
    "path"
] = "s3a://superconductive-docs-test/data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01.csv"

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)

batch_list: List[Batch] = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 1

batch: Batch = batch_list[0]
assert batch.data.dataframe.shape[0] == 10000

# Here is a BatchRequest naming a data_asset
batch_request = BatchRequest(
    datasource_name="my_s3_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name = (
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01"
)

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["my_s3_datasource"]
assert set(
    context.get_available_data_asset_names()["my_s3_datasource"][
        "default_inferred_data_connector_name"
    ]
) == {
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01",
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-02",
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-03",
}

batch_list: List[Batch] = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 1

batch: Batch = batch_list[0]
assert batch.data.dataframe.shape[0] == 10000



"""POC for loading config."""

from __future__ import annotations

import logging
import pathlib
from pprint import pformat as pf
from typing import TYPE_CHECKING, Dict, List, Type, Union

from typing_extensions import Final

from great_expectations.compatibility.pydantic import (
    Extra,
    Field,
    ValidationError,
    validator,
)
from great_expectations.experimental.datasources.experimental_base_model import (
    ExperimentalBaseModel,
)
from great_expectations.experimental.datasources.interfaces import (
    Datasource,  # noqa: TCH001
)
from great_expectations.experimental.datasources.sources import _SourceFactories

if TYPE_CHECKING:
    from great_expectations.compatibility.pydantic.error_wrappers import (
        ErrorDict as PydanticErrorDict,
    )


logger = logging.getLogger(__name__)


_ZEP_STYLE_DESCRIPTION: Final[str] = "ZEP Experimental Datasources"

_MISSING_XDATASOURCES_ERRORS: Final[List[PydanticErrorDict]] = [
    {
        "loc": ("xdatasources",),
        "msg": "field required",
        "type": "value_error.missing",
    }
]


class GxConfig(ExperimentalBaseModel):
    """Represents the full new-style/experimental configuration file."""

    xdatasources: Dict[str, Datasource] = Field(..., description=_ZEP_STYLE_DESCRIPTION)

    @property
    def datasources(self) -> Dict[str, Datasource]:
        return self.xdatasources

    class Config:
        extra = Extra.ignore  # ignore any old style config keys

    @validator("xdatasources", pre=True)
    @classmethod
    def _load_datasource_subtype(cls, v: Dict[str, dict]):
        logger.info(f"Loading 'datasources' ->\n{pf(v, depth=2)}")
        loaded_datasources: Dict[str, Datasource] = {}

        for ds_name, config in v.items():
            ds_type_name: str = config.get("type", "")
            if not ds_type_name:
                # TODO: (kilo59 122222) ideally this would be raised by `Datasource` validation
                # https://github.com/pydantic/pydantic/issues/734
                raise ValueError(f"'{ds_name}' is missing a 'type' entry")

            try:
                ds_type: Type[Datasource] = _SourceFactories.type_lookup[ds_type_name]
                logger.debug(f"Instantiating '{ds_name}' as {ds_type}")
            except KeyError as type_lookup_err:
                raise ValueError(
                    f"'{ds_name}' has unsupported 'type' - {type_lookup_err}"
                ) from type_lookup_err

            datasource = ds_type(**config)
            loaded_datasources[datasource.name] = datasource

            # TODO: move this to a different 'validator' method
            # attach the datasource to the nested assets, avoiding recursion errors
            for asset in datasource.assets.values():
                asset._datasource = datasource

        logger.info(f"Loaded 'datasources' ->\n{repr(loaded_datasources)}")

        if v and not loaded_datasources:
            raise ValueError(f"Of {len(v)} entries, no 'datasources' could be loaded")
        return loaded_datasources

    @classmethod
    def parse_yaml(
        cls: Type[GxConfig], f: Union[pathlib.Path, str], _allow_empty: bool = False
    ) -> GxConfig:
        """
        Overriding base method to allow an empty/missing `xdatasources` field.
        Other validation errors will still result in an error.

        TODO (kilo59) 122822: remove this as soon as it's no longer needed. Such as when
        we use a new `config_version` instead of `xdatasources` key.
        """
        if _allow_empty:
            try:
                super().parse_yaml(f)
            except ValidationError as validation_err:
                errors_list: List[PydanticErrorDict] = validation_err.errors()
                logger.info(
                    f"{cls.__name__}.parse_yaml() failed with errors - {errors_list}"
                )
                if errors_list == _MISSING_XDATASOURCES_ERRORS:
                    logger.info(
                        f"{cls.__name__}.parse_yaml() returning empty `xdatasources`"
                    )
                    return cls(xdatasources={})
                else:
                    logger.warning(
                        "`_allow_empty` does not prevent unrelated validation errors"
                    )
                    raise
        return super().parse_yaml(f)
